# encoding: utf-8
from yade import qt, export, pack, plot
from yade.gridpfacet import *
import time

start_time = time.time()

# This script is built from the tutorial yade script developed by Bruno Chareyre (2010)
# Found at https://gitlab.com/yade-dev/trunk/-/blob/master/examples/triax-tutorial/script-session1.py

###################################
##### User defined parameters #####
###################################

clumpSample = 1  # generate sample of clumps?

targetPorosity = 0.4  # target porosity of the REFD algorithm
porosityLoop=0  # REFD porosity algorithm on or off

# sample dimensions
widthX=0.1
widthY=0.02   # shear loc -> 0.02
height=0.25 #height of cylinder/cube

timestepSafety = .8  #default 0.8
damping = 0.03   #default 0.1
sampleToWidth = 1 # 0.99 originally to allow the membrane to not start in contact with sample.
mn,mx=Vector3((1-sampleToWidth)*widthX,(1-sampleToWidth)*widthY,0),Vector3(widthX*sampleToWidth,widthY*sampleToWidth,height) # corners of the initial packing

gravity = (0,0,0)

# Isotropic stress
sigmaIso =-200e3  # test -20e3 OK

# Anisotropic stress goals in x,y,z directions
sigmaX=sigmaIso * 1
sigmaY=sigmaIso
sigmaZ=sigmaIso * 1

stabilityThreshold=0.03  #test unbalancedForce against this value in different loops (see below)
if widthY!=widthX:
    shallowSample=1
if sigmaX != sigmaY:
    stabilityThreshold=0.1
    stabilityThresholdDev=0.05
iterPerCheck = 500  # Check if stable during compaction phase

# for spheres and plates
young = 200e6  # default 300e6
youngWalls=1e10
poisson = 0.3
density = 2800
compFricDegree = 15 # compaction friction angle -> determines the porosity

rMean= .0011  # mean radius
rRelFuzz= .4  # radius variability
num_spheres=-1  # number of spheres. -1 generates the maximum possible within the boundaries

###################################
##### Materials #####
###################################
# Create materials for spheres and plates
O.materials.append(FrictMat(young=young, poisson=poisson, frictionAngle=radians(compFricDegree), density=density,
                            label='spheres'))

#### ENFORCE ZERO FRICTION on all loading plates here:
O.materials.append(FrictMat(young=youngWalls, poisson=poisson, frictionAngle=0, density=0,
                            label='compWalls'))

#### friction on top and bot walls
wallZMatId=O.materials.append(FrictMat(young=youngWalls, poisson=poisson, frictionAngle=radians(compFricDegree), density=0,
                            label='ZCompWalls'))

oversizeFactor=sigmaY/sigmaX*10
if sigmaX==sigmaY==sigmaZ:  # isotropic stress -> smaller load plates necessary
    oversizeFactor=2

## create walls around the packing
walls=aabbWalls([mn,mx],thickness=0,oversizeFactor=oversizeFactor,material='compWalls')
wallIds=O.bodies.append(walls)


# assign friction on top and bot walls
#top
walls[5].mat=O.materials[wallZMatId]
#bot
walls[4].mat=O.materials[wallZMatId]


###########################
##### ENGINES #####
###########################

triax=TriaxialStressController(  # controls stress and strain
    maxMultiplier=1.+2e5/young, # spheres growing factor (fast growth)
    finalMaxMultiplier=1.+2e4/young, # spheres growing factor (slow growth)
    ## DEFAULT:
    # maxMultiplier=1. + 2e4 / young,  # spheres growing factor (fast growth)
    # finalMaxMultiplier=1. + 2e3 / young,  # spheres growing factor (slow growth)
    thickness = 0,
    stressMask = 7, # binary: 0111 -> stress controlled in all dir
    internalCompaction=1, # If true the confining pressure is generated by growing particles
)

O.engines=[
    ForceResetter(),
    InsertionSortCollider([Bo1_Sphere_Aabb(), Bo1_Box_Aabb()]),
    InteractionLoop(
        [Ig2_Sphere_Sphere_ScGeom6D(), Ig2_Box_Sphere_ScGeom()], #Ig2_Wall_Sphere_ScGeom(),
        [Ip2_FrictMat_FrictMat_FrictPhys()],
        [Law2_ScGeom_FrictPhys_CundallStrack()]
    ),
    GlobalStiffnessTimeStepper(active=1, timeStepUpdateInterval=100,
                               timestepSafetyCoefficient=timestepSafety,label='ts'),
        # GlobalStiffnessTimeStepper(active=1,timeStepUpdateInterval=100,
    triax,
    NewtonIntegrator(gravity=gravity,damping=damping,label='newton'),
    #PyRunner(command='checkUnbalanced()',realPeriod=1,label='checker')  # Check if stable
]

##########################
###  Support functions ###
##########################

def samplePorosity(res):
    facetXMin = O.bodies[wallIds[0]].state.pos[0]
    facetXMax = O.bodies[wallIds[1]].state.pos[0]
    facetYMin = O.bodies[wallIds[2]].state.pos[1]
    facetYMax = O.bodies[wallIds[3]].state.pos[1]
    facetZMin = O.bodies[wallIds[4]].state.pos[2]
    facetZMax = O.bodies[wallIds[5]].state.pos[2]
    startPoint = Vector3(facetXMin, facetYMin, facetZMin)
    endPoint = Vector3(facetXMax, facetYMax, facetZMax)

    # end and start define the corners of the cubic volume to check porosity of
    return voxelPorosity(resolution = res, start = startPoint, end = endPoint )

###################################
##### Generate sample #####
###################################

# create empty sphere packing
sp=pack.SpherePack()
# generate randomly spheres with uniform radius distribution
sp.makeCloud(mn,mx,rMean,rRelFuzz, num_spheres, 0,0.7,seed=1)
sp.toSimulation(material='spheres')

numParticles = len(O.bodies)-6 # minus the 6 facets
print ("Number of particles: ",numParticles)

if clumpSample:
    # clump type 1
    relRadList1 = [1, 1]                       # radii of clump members
    relPosList1 = [[.6, 0, 0], [-.6, 0, 0]]    # positions of clump members

    templates = []
    templates.append(clumpTemplate(relRadii=relRadList1, relPositions=relPosList1))
    # replace spheres with clumps
    clumpIds= O.bodies.replaceByClumps(templates, [1])  # returns clumpId and memberIds
                                                                # -> skip memberIds with _
    # example with 3 different clump templates at 50% 30% and 10% of sample + leave 10% of original spheres as is
    # clumpIds, _ = O.bodies.replaceByClumps(templates, [.5, .3, .1])

    O.bodies.updateClumpProperties()  # update to get correct volume, mass, inertia of clumps

if clumpSample:
    part = O.bodies[20+2*numParticles]
    startRad = part.shape.radius
else:
    part = O.bodies[20]
    startRad = part.shape.radius

########################
#### Calm Function ####
########################
# # stabilize sample when using replaceByClumps -> swapping in clumps can lead to overlaps between spheres
# if clumpSample:
# 	print('\nPlease wait a minute ...\n')
# 	O.engines=O.engines+[PyRunner(iterPeriod=100,command='calm()',label='calmRunner')]
# 	O.run(10000,True)  # default 1000000   evt use young??
# 	calmRunner.dead=True

#######################################
###  Isotropic Confining Pressure  ###
#######################################
# the value of (isotropic) confining stress defines the target stress to be applied in all three directions
triax.goal1=triax.goal2=triax.goal3=sigmaIso

while 1:
  O.run(iterPerCheck, 1)
  #the global unbalanced force on dynamic bodies, thus excluding boundaries, which are not at equilibrium
  unb=unbalancedForce()
  print('unbalanced force:',unb,' mean stress: ',triax.meanStress)
  if unb<stabilityThreshold and abs(sigmaIso-triax.meanStress)/(-sigmaIso)<0.01:
    break

#######################################
###  Anisotropic Conf Pressure  ###
#######################################
if not sigmaX==sigmaY==sigmaZ:
    # at this stage the sample dimensions will change in order to allow for anistropic stress

    # turn internal compaction off to keep constant particle sizes
    triax.internalCompaction=0

    # reduce timestep
    timestepSafety=.08
    ts.timestepSafetyCoefficient=timestepSafety

    # set new stress goals
    triax.goal1=sigmaX
    triax.goal2=sigmaY
    triax.goal3=sigmaZ

    if not clumpSample:
        stabilityThreshold=0.15

    while 1:
        O.run(iterPerCheck, 1)
        #the global unbalanced force on dynamic bodies, thus excluding boundaries, which are not at equilibrium
        unb=unbalancedForce()
        stressX=0.5*(triax.stress(0)[0]+triax.stress(1)[0])
        stressY=0.5*(triax.stress(2)[1]+triax.stress(3)[1])
        stressZ=0.5*(triax.stress(4)[2]+triax.stress(5)[2])
        print("stressX:", stressX, "\tstressY:", stressY, "\tstressZ:", stressZ, " \tunb:",unb)
        #print("X: ",abs((stressX-sigmaX)/sigmaX) ," Y: ",abs((stressY-sigmaY)/sigmaY)," Z:",abs((stressZ-sigmaZ)/sigmaZ))
        if unb<stabilityThresholdDev and abs((stressX-sigmaX)/sigmaX)<0.01 and \
                abs((stressY-sigmaY)/sigmaY)<0.01 and abs((stressZ-sigmaZ)/sigmaZ)<0.01:
            break

###################################################
###   REACHING A SPECIFIED POROSITY PRECISELY   ###
###################################################

### We will reach a prescribed value of porosity with the REFD algorithm
### (see http://dx.doi.org/10.2516/ogst/2012032 and
### http://www.geosyntheticssociety.org/Resources/Archive/GI/src/V9I2/GI-V9-N2-Paper1.pdf)
print("Porosity at istr. comp. stress: ",triax.porosity)
if porosityLoop:
    while samplePorosity(300)>targetPorosity:
        # we decrease friction value and apply it to all the bodies and contacts
        compFricDegree = 0.90*compFricDegree
        setContactFriction(radians(compFricDegree))
        print "\r Friction: ",compFricDegree," porosity:",triax.porosity,
        sys.stdout.flush()
        # while we run steps, triax will tend to grow particles as the packing
        # keeps shrinking as a consequence of decreasing friction. Consequently
        # porosity will decrease
        O.run(500,1)
    print("Porosity after ",triax.porosity)

# pause to set final positions
O.pause()

# 3D view of compacted sample
qt.View()


####################
###  Tesselation ###
####################
# TW=TesselationWrapper()
# TW.triangulate()        #compute regular Delaunay triangulation, don’t construct tesselation
# TW.computeVolumes()     #will silently tesselate the packing, then compute volume of each Voronoi cell
# TW.setState(0)
# TW.saveState(state=0, bz2=True)
# #TW.setState(0)          #store current positions internaly for later use as the "0" state
# #
# # O.run(100,True)         #make particles move a little (let's hope they will!)
# # TW.setState(1)          #store current positions internaly in the "1" (deformed) state
# # #Now we can define strain by comparing states 0 and 1, and average them at the particles scale
# # TW.defToVtk("strain.vtk")


####################
###  Save Data ###
####################
# get final sample dimensions:
facetXMin=O.bodies[wallIds[0]].state.pos[0]
facetXMax=O.bodies[wallIds[1]].state.pos[0]
facetYMin = O.bodies[wallIds[2]].state.pos[1]
facetYMax = O.bodies[wallIds[3]].state.pos[1]
facetZMin = O.bodies[wallIds[4]].state.pos[2]
facetZMax = O.bodies[wallIds[5]].state.pos[2]
wallFinalPos=[facetXMin, facetXMax, facetYMin, facetYMax, facetZMin, facetZMax]

# save the positions and radii of spheres/clumps
if clumpSample:
    export.textClumps('compactSample.txt')
else:
    export.text('compactSample.txt')
print ("Sample saved")
print('Elapsed time in minutes: ' + str((time.time() - start_time) / 60))
samplePor = samplePorosity(600)
print("Sample porosity: "+str(samplePor))

# Find radius scaling factor from sample gen to final compacted sample
print("internalCompaction -> endRad/startRad = " + str(part.shape.radius/startRad))


# save other parameters to txt file
f = open("1paramSampleGen.txt", "w")  # mode "w" opens a new file and will erase an existing file
# mode "a+" creates a new file or appends to an existing file
f.write("wallFinalPos="+str(wallFinalPos)+"\n")
f.write("clumpSample="+str(clumpSample)+"\n")
f.write("shallowSample="+str(shallowSample)+"\n")
f.write("numParticles="+str(numParticles)+"\nwidthX="+str(widthX)+"\nwidthY="+str(widthY)+"\nheight="+str(height)+"\n")
f.write("sigmaIso="+str(sigmaIso)+"\n"+"sigmaX="+str(sigmaX)+"\nsigmaY="+str(sigmaY)+"\nsigmaZ="+str(sigmaZ)+"\n")

f.write("youngWalls="+str(youngWalls)+"\nyoung="+str(young)+"\npoisson="+str(poisson)+"\ndensity="+str(density)+"\n")
f.write("compFricDegree="+str(compFricDegree)+"\nrMean="+str(rMean)+"\nrRelFuzz="+str(rRelFuzz)+"\n")
f.write("damping="+str(damping)+"\n")

f.write("compactedPorosity="+str(samplePor)+"\n")
f.close()




